 #include "Mainwindow.hxx"
#include "widgets/BasicBlockWidget.hxx"
#include "qt.hxx"
#include "disassembler/llvm/LLVMDisassembler.hxx"

#include <iostream>
#include <sstream>

#include <QtGui>

Mainwindow::Mainwindow(const std::string& filename)
{
    openAction = new QAction(tr("&Open"), this);
    // saveAction = new QAction(tr("&Save"), this);
    exitAction = new QAction(tr("E&xit"), this);

    connect(openAction, SIGNAL(triggered()), this, SLOT(open()));
    // connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));
    connect(exitAction, SIGNAL(triggered()), qApp, SLOT(quit()));

    fileMenu = menuBar()->addMenu(tr("&File"));
    fileMenu->addAction(openAction);
    // fileMenu->addAction(saveAction);
    fileMenu->addSeparator();
    fileMenu->addAction(exitAction);

    listWidget = new QListWidget();
    stackedWidget = new QStackedWidget();
    dockWidget = new QDockWidget(tr("Functions"), this);
    dockWidget->setAllowedAreas(Qt::LeftDockWidgetArea |
                                Qt::RightDockWidgetArea);
    dockWidget->setWidget(listWidget);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget);
    setCentralWidget(stackedWidget);

    connect(listWidget, SIGNAL(currentRowChanged(int)),
            stackedWidget, SLOT(setCurrentIndex(int)));

    setWindowTitle(tr("FRIDA"));

	openBinary(filename);
}

void Mainwindow::quit()
{
    QMessageBox messageBox;
    messageBox.setWindowTitle(tr("Notepad"));
    messageBox.setText(tr("Do you really want to quit?"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);
    if (messageBox.exec() == QMessageBox::Yes)
        qApp->quit();
}

void Mainwindow::open() {
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), "",
                                                    tr("Binaries (*)"));

	openBinary(fileName.toStdString());
}

void Mainwindow::openBinary(const std::string& filename) {
    if (filename != "") {
        disassembler.reset(new LLVMDisassembler(filename));
		disassembler->forEachFunction([&](uint64_t address, Function* fun) {
				populateSymbolInformation(fun);
			});
    }
}

namespace {
	void local__add_basic_block(BasicBlock * block, Disassembler * dis, QGraphicsScene * scene,
								uint64_t starty, uint64_t startx) {
		std::stringstream s;
		s << "BLOCK_" << std::hex << block->getStartAddress()
		  << "_" << block->getEndAddress();
		BasicBlockWidget * widget = new BasicBlockWidget(s.str().c_str());
		scene->addItem(widget);
		widget->setFlag(QGraphicsItem::ItemIsMovable, true);
		widget->moveBy(100*startx, 10*(block->getStartAddress() - starty));

		dis->printEachInstruction(block->getStartAddress(), block->getEndAddress(),
								  [&](uint8_t* bytes, size_t byte_count, const std::string& line) {
									  widget->addItem(bytes, byte_count, line.c_str() + 1);
								  });

		if (block->getNextBlock(0) != 0)
			local__add_basic_block(dis->getBasicBlock(block->getNextBlock(0)), dis, scene, starty, startx+1);
		if (block->getNextBlock(1) != 0)
			local__add_basic_block(dis->getBasicBlock(block->getNextBlock(1)), dis, scene, starty, startx-1);
	}
}

void Mainwindow::populateSymbolInformation(Function* fun) {
    QTabWidget * w = new QTabWidget();

    // CFG
    QGraphicsScene * scene = new QGraphicsScene;

	BasicBlock * block = disassembler->getBasicBlock(fun->getStartAddress());

	local__add_basic_block(block, disassembler.get(), scene, block->getStartAddress(), 100);

    QGraphicsView * view = new QGraphicsView(scene);
    w->addTab(view, "CFG");

    // Listing
    QTableWidget * t = new QTableWidget();
    t->setColumnCount(3);
    t->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);

    w->addTab(t, "Listing");

    listWidget->addItem(fun->getName().c_str());
    stackedWidget->addWidget(w);
}
